#!/usr/bin/perl
# --------------------------------------------------------------------
# nproc - Nodediffs Processor, version 1.1
# (C)opyright by val khokhlov
#
use File::Copy;
use POSIX qw(strftime);
#
# Nodelist and nodediffs path (absolute path)
$NLIST_PATH  = "/home/val/fido/nodelist";
$NDIFF_PATH  = "/home/val/fido/in.sec";
#
# Temporary base directory (absolute path)
$TMP_PATH   = "/tmp";
#
# How to get original directory (`pwd` for unix; `cd` for dos)
$ORIG_PATH = `pwd`; chomp $ORIG_PATH;
#
# Lists to process, list=>'base_name', diff=>'diff_name'
@NODELISTS  = (
                 { list=>'nodelist', diff=>'NODEDIFF' },
		 { list=>'pnt46reg', diff=>'R46PDIFF' }
 	      );
#
# Unpackers, 'ext_letter'=>'command to execute'
%UNPACKERS  = (
                'z'=>'/usr/local/bin/unzip -j -o -L "%n.Z??"',
                'r'=>'/usr/local/bin/rar e -y -cl "%n.R??"'
#                'z'=>'c:\\arc\\pkunzip.exe %n.z??'
	      );
#
# CRC check - either 1 (internal module), or program name or undefined
# note: program - with full path!, skip check if undefined
#$CHECK_CRC = "$ORIG_PATH/crc";
#$CHECK_CRC = '/home/val/fido/nodelist/crc';
$CHECK_CRC = 1;
#
# Delete old lists and diffs from original directories
$DEL_OLDLISTS = 0;
$DEL_OLDDIFFS = 0;
#
# Time to ignore another copy in hours (0 = always, -1 = never)
$IGNORE_TIME  = 1;
#
# Log file name (relative to run dir or absolute), or no log if commented
$LOG_NAME     = "/home/val/fido/nodelist/nproc.log";
#
# Display output to stdout if non-zero
$VERBOSE      = 0;

# lprint
sub lprint
{
  my $s = strftime("%d.%m.%y %H:%M ", localtime);
  foreach (@_) { $s .= $_ };
  if ( defined $LOG_NAME ) { print LOG $s; }
  if ( $VERBOSE == 1 ) { print $s; }
}

# nproc(nodelist, nodediff): apply diff to list, return new list name or ''
sub nproc()
{
  my $i;
  $/ = "\r\n";
  $nlist = shift; 
  $ndiff = shift;
  # check nodelist file extension
  if (!($nlist =~ m!([^\/\\]*\.)(\d{3})$!)) { lprint "Warning: Bad list name $nlist\n"; return; };
  $oeday = $2; $newlist = $1;
  # check nodediff file extension
  if (!($ndiff =~ m!\.(\d{3})$!)) { lprint "Warning: Bad diff name $ndiff\n"; return; };
  $neday = $1; $newlist .= $1;
  # open files
  open L, "$nlist"; binmode L;
  open D, "$ndiff"; binmode D;
  # read first line of nodelist and compare days
  $o1 = $_ = <L>;
  if (!m/;.*?Day number (\d+) : (\d{5})/) { lprint "Warning: Bad list $nlist\n"; close L; close D; return; }
  $ocday = $1; $ocrc = $2;
  if ($oeday != $ocday) { lprint "Warning: $nlist days aren't same: ext=$oeday, file=$ocday\n"; close L; close D; return; }
  # read first line of nodediff and compare to that of nodelist
  $n1 = $_ = <D>;
#  if (!m/;.*?Day number (\d+) : (\d{5})/) { warn "Bad nodediff $ndiff;"; return; }
#  $ncday = $1; $ncrc = $2;
#  if ($neday != $ncday) { warn "Nodediff $ndiff days not same: ext=$neday, file=$ncday"; return; }
  if ($o1 ne $n1) { lprint "Warning: First line of $ndiff doesn't match first line of $nlist\n"; close L; close D; return; }
  # create new nodelist and process
  lprint "Compiling: $nlist + $ndiff -> $newlist\n";
  open N, ">$newlist"; binmode N;
  $n = 0; $err = 0; $first_line = 1; $rcrc = 0;
  while (<D>) {
    $n++;
    chomp $_;
    if (!m/^([ACD])(\d+)$/) { lprint "Warning: Bad line #$n in $ndiff: $_\n"; $err = 1; last; }
    $cmd = $1; $cnt = $2;
    if ($n == 1 && (($1 eq 'C') || ($1 eq 'D'))) { $cnt--; }
#    if ($cmd eq 'A') { for ($i = 0; $i < $cnt; $i++) { $_ = <D>; $n++; print N $_; } }
    if ($cmd eq 'A') {
      for ($i = 0; $i < $cnt; $i++) { 
        $_ = <D>; 
        $n++; print N $_; 
        if ($first_line) { ($crc) = m!(\d{5})$/!; $first_line = 0; }
        elsif ($CHECK_CRC) { $rcrc = crc16::crc16($_, length $_, $rcrc); }
      }
    }
    elsif ($cmd eq 'C') { 
      for ($i = 0; $i < $cnt; $i++) { 
        $_ = <L>; print N $_; 
        if ($CHECK_CRC) { $rcrc = crc16::crc16($_, length $_, $rcrc); }
      } 
    }
    elsif ($cmd eq 'D') { for ($i = 0; $i < $cnt; $i++) { $_ = <L>; } }
  }
  # close files
  print N "\x1a";
  close N; close D; close L;
  # check crc
  if (defined $CHECK_CRC) { 
    $rcrc = `$CHECK_CRC $newlist` unless $CHECK_CRC;
    if ($crc != $rcrc) { lprint "Warning: CRC mismatch (in diff $crc, calculated $rcrc)\n"; $err = 1; }
  }
  # if error - delete new nodelist
  if ($err) { unlink $newlist; return; }
  return $newlist;
}
# nfind(nodelist-name, nodediff-name)
sub nfind()
{
  my $i;
  $list = shift;
  $diff = shift;
  # copy files
  @lists = glob("$NLIST_PATH/$list.*");
  @diffs = glob("$NDIFF_PATH/$diff.*");
  @files = (@lists, @diffs);
  foreach (@files) { copy $_, "./"; }
  # unpack
  while ( ($ext, $_) = each %UNPACKERS ) { s/%n/$diff/; `$_`; }
  # get max nodelist number
  $max = 0; $year = 0;
  foreach (@lists) {
    if (!m/\.(\d{3})$/) { next; }
    $y = (localtime((stat $_)[9]))[5]+1900;
    if ( ($1 > $max) && ($y >= $year) ) { $max = $1; $year = $y; $clist = $_; }
  }
  $clist =~ s!$NLIST_PATH/!!;
  # get all names of diffs that are greater, than nodelist
  $diff =~ tr/[A-Z]/[a-z]/;
  @files = glob("$diff.*");
  @diffs_sorted = ();
  foreach (@files) {
    if (!m/\.(\d{3})$/) { next; }
    $y = (localtime((stat $_)[9]))[5]+1900;
    if ( ($1 <= $max) && ($y <= $year) ) { next; }
    for ($i = 0; $i <= @diffs_sorted; $i++) {
      if ( $i == @diffs_sorted ) { splice @diffs_sorted, $i, 0, {'n'=>$1, 'y'=>$y, 'f'=>$_}; last; }
      elsif ( $1 < $diffs_sorted[$i]{'n'} && $y <= $diffs_sorted[$i]{'y'} ) { splice @diffs_sorted, $i, 0, {'n'=>$1, 'y'=>$y, 'f'=>$_}; last; }
    }
  }
  # compile nodelist
  if (defined $CHECK_CRC) { $crc_str = " (CRC Ok)"; }
  @diffs = ();
  foreach (@diffs_sorted) {
    my $num = $_->{'n'};
    #lprint "Nodediff: ".$_->{'f'}."; number ".$_->{'n'}."; year ".$_->{'y'}."\n";
    $s = &nproc($clist, $_->{'f'});
    if ($s) { 
      lprint "Compiled: $s".$crc_str."\n"; $clist = $s; 
      # add processed diff-file to to-delete list
      push @diffs, glob $NDIFF_PATH.'/'.$diff.'.?'.($num % 100); 
    }
  }
  for ($i = 0; $i < @lists; $i++) { undef $lists[$i] if $clist eq $lists[$i]; }
  # remove old files and move new nodelist to nodelist dir
  if (!-e $NLIST_PATH.'/'.$clist) {
    if (move $clist, "$NLIST_PATH/") {
      unlink @lists if $DEL_OLDLISTS;
      unlink @diffs if $DEL_OLDDIFFS;
      $cyday = (localtime)[7]+1; ($nyday) = $clist =~ /\.(\d{3}$)/;
      if ($cyday - $nyday < 7) { $old = "up to date"; }
        else { $old = (int(($cyday-$nyday)/7))." week(s) old"; }
      lprint "All done with $list, final list is $clist ($old)\n";
    }
    else { lprint "Can't move $clist to $NLIST_PATH, compilation reverted\n"; }
  }
  else { lprint "No updates for $clist\n"; }
  unlink glob("$TMP_PATH/nproc/*");
}

# open log file
if ( defined $LOG_NAME ) { open LOG, ">>$LOG_NAME"; }
# check if we can execute crc
if (defined $CHECK_CRC) {
  if ($CHECK_CRC) {
    my $test = eval { require crc16; 1 };
    if ( !$test ) { lprint "Error: Can't load crc16 perl module or library\n$@"; exit 1; }
    $_ = 'val khokhlov';
    $test = crc16::crc16($_, length $_, 0);
    if ($test != 0x913b) { lprint "Error: Internal crc16 module check failure\n"; exit 1; }
  }
  elsif (! -x $CHECK_CRC) { lprint "Error: Can't find executable file $CHECK_CRC\n"; exit 1; }
}
# create temp dir
if (-d "$TMP_PATH/nproc") {
  $hours = $IGNORE_TIME / 24;
  if (($hours < 0) || (-M "$TMP_PATH/nproc" < $hours)) { lprint "Error: Another copy of nproc may be active! Fresh $TMP_PATH/nproc exist!\n"; exit 1; }
  lprint "Warning: $TMP_PATH/nproc exists, removing all contents!\n";
  unlink glob("$TMP_PATH/nproc/*");
}
else { mkdir "$TMP_PATH/nproc", 0775; }
chdir "$TMP_PATH/nproc";
# process nodelists
for ($i = 0; $i < @NODELISTS; $i++) {
  &nfind($NODELISTS[$i]{'list'}, $NODELISTS[$i]{'diff'});
}
# return
if (!rmdir("$TMP_PATH/nproc")) { lprint "Warning: Can't remove temp dir $TMP_PATH/nproc ($!)\n"; }
chdir $ORIG_PATH;
if ( defined $LOG_NAME ) { print LOG "\n"; close LOG; }
